// PdfAppUi.h
//
// Copyright (c) 1999-2000 Sander van der Wal.  All rights reserved.
//
// $Id: pdfappui.h 1.1 2000-09-17 13:18:10+02 svdwal Exp svdwal $

#ifndef PDFAPPUI_H_
#define PDFAPPUI_H_

// Parents
#ifndef __EIKAPPUI_H__
#include <eikappui.h>
#endif
#include "PdfPageView.h"
#include "SearchState.h"

// implementation details
#ifndef __BADESCA_H__
#include <badesca.h>
#endif


//////////////////////////////////////////////////////////////////////////////
//
// TPdfUiState
//
//////////////////////////////////////////////////////////////////////////////
struct TPdfUiState: public TPdfPageViewMemento
{
public:
  TPdfUiState();

  void SetDefaults();
  
  TBool operator!=(const TPdfUiState& y);

  void InternalizeL(RReadStream& aStream);
  void ExternalizeL(RWriteStream& aStream) const;

  void StoreL(CDictionaryStore* aIniFile, TUid aStreamUid) const;
  void RestoreL(CDictionaryStore* aIniFile, TUid aStreamUid);
 
public:
 
  TInt  iHBarVisibility;
  TInt  iVBarVisibility;
 
  TBool iShowToolBar;
  TBool iShowToolBand;
  TBool iMaximizeView;
};

struct SEdwinFindModel;
class CEikCommandButtonBase;

class CPdfTextListBox;
class CSearchState;


//////////////////////////////////////////////////////////////////////////////
//
// CPdfAppUi definition
//
//////////////////////////////////////////////////////////////////////////////
class CPdfAppUi: public CEikAppUi,
                 public MCoeControlObserver, 
                 public MPdfPageViewCmdHandler
{
public:
  void ConstructL();
  ~CPdfAppUi();
 	
// --o CEikAppUi
protected:
  void HandleCommandL(TInt aCommand);

  TBool ProcessCommandParametersL(TApaCommand aCommand,
                                  TFileName& aDocumentName,
                                  const TDesC& aTail);
  void HandleModelChangeL();
  void DynInitMenuPaneL(int aMenuId, CEikMenuPane *aMenuPane);
  void OpenFileL(const TDesC& aFileName);
  TErrorHandlerResponse HandleError(TInt aError, const SExtendedError& aExtErr, TDes& aErrorText, TDes& aContextText);

  // --o MCoeControlObserver
private:
  virtual void HandleControlEventL(CCoeControl* aControl,TCoeEvent aEventType);

// --o CPdfAppUi
protected:
  static const TUid KPreferencesStreamUid;
  void ChangePageL(TInt aPage, TBool aTop=ETrue);
  
  virtual void SetUiStateForModelL();

  void LoadUiStateL(const TUid& aStreamUid, TPdfUiState& aUiState);

private:
  void SetModel();
  void UpdateFileNameLabelL();
  TFileName CurrentFilePath() const;
  void SetInitialPathL(TFileName& aFileName) const;
 
  void StartRenderingL();           // current page
  void SetPage(TInt aPage);
  void SetToolBarsL(TBool aShowToolBar, TBool aShowToolBand);
  void SetDrawAccurate(TBool aDrawAccurate);
  void SetZoom(TInt aZoom);
  void DoZoomL(TInt aZoom);
  void SetMaximizeViewL(TBool aMaximizeView);
  
  void ReflectUiStateL();
  void UpdateUiStateL(const TPdfUiState& aNewState);
  void StoreUiStateL(const TUid& aStreamUid, const TPdfUiState& aUiState) const;

  void ZoomPageL(TInt aZoom);
  void DoHelpContentsL(TBool aStartup);
  
  void DoSearch();
  void DoCreatePopoutL(CDesCArray* aText, TInt aIndex, CEikCommandButtonBase* aLaunchingButton);

  // Commands
  void CmdCancelL();

  // file menu
  void CmdFileOpenL();
  void CmdDocumentInfoL();
  void CmdExportAsTextL();
  void CmdExit();

  // find menu
  void CmdSearchFindL();
  void CmdSearchFindAgainL();
  void CmdGoToL();
  
  // view menu
  void CmdZoomInL();
  void CmdZoomOutL();
  void CmdSetViewL();
  void CmdShowToolBarsL(TInt aCmd);
  void CmdMaximizeViewL();

  // tools menu
  void CmdPreferencesL();
 
#ifdef TRACE
  void CmdShowProfile();
#endif

#ifdef OOM
  void CmdOOML();
  PDFDoc* DoOpenFileOOML(const TFileName& aFileName);
  void DisplayPageOOML(TInt aPage);
#endif
  void CmdHelpAboutL();

  // button commands
  void CmdZoomPopupL();
  void CmdFitActualSizeL();
  void CmdFitPageL();
  void CmdFitWidthL();
  
  void CmdFirstPageL();
  void CmdPrevPageL(TBool aTop = ETrue);
  void CmdNextPageL();
  void CmdLastPageL(TBool aTop = ETrue);

  void CmdDrawAccurateL();

  // misc
  void CmdTapL(TPoint /* aPos */) {}

protected:
  PDFDoc* iDoc;

  // user interface elements
  CPdfPageView* iPageView;
  CCoeControl*  iAppView;

  // UI state
  TInt iPage;
  TPdfUiState  iUiState; // this is saved
  
  // Search state
  SEdwinFindModel* iFindModel;
  CDesCArray*      iFindList;
  CSearchState*    iSearchState;
  TRect            iSelection; 

  // popup boxes
  CPdfTextListBox* iPopoutListBox;
  TBool            iZoomBoxPopped;
};

#endif // PDFAPPUI_H_
