// PdfPageView.h
//
// Copyright (c) 1999 Sander van der Wal.  All rights reserved.
//

#ifndef PDFPAGEVIEW_H_
#define PDFPAGEVIEW_H_

// $Id: pdfpageview.h 1.1 2000-09-17 13:18:13+02 svdwal Exp svdwal $

// Parents
#ifndef __COECNTRL_H__
#include <coecntrl.h>  // CCoeControl
#endif
#ifndef __COECCNTX_H__
#include <coeccntx.h>  // MCoeControlBrushContext
#endif
#ifndef __EIKSCRLB_H__
#include <eikscrlb.h>  // MEikScrollbarObserver
#endif


// --o Files
class RWriteStream;

// --o Eikon
class CEikBusyMsgWin;

// --o PDF
#include "PdfPageControl.h"

class FontEncoding;
class PDFDoc;

//////////////////////////////////////////////////////////////////////////////
//
// TPdfPageViewLayout
//
// Calculates and contains the positions and sizes of the child control 
// and scrollbars
//
//////////////////////////////////////////////////////////////////////////////
class TPdfPageViewLayout
{
public:
  TPdfPageViewLayout();
  TPdfPageViewLayout(const TRect& aParentRect, const TSize& aChildSize, TInt aShowHBar, TInt aShowVBar);
  // copy constructor
  TPdfPageViewLayout(const TPdfPageViewLayout& aLayout);

// --o TPdfPageViewLayout
public:
  inline TRect ControlRect() const { return iControlRect;};
  
  inline TBool ShowHBar() const { return iShowHBar; };
  inline TPoint HBarPos() const { return iHBarPos; };
  inline TInt HBarLength() const { return iHBarLength; };

  inline TBool ShowVBar() const { return iShowVBar; };
  inline TPoint VBarPos() const { return iVBarPos; };
  inline TInt VBarLength() const { return iVBarLength; };

  inline TInt VisAreaWidth() const { return iVisArea.iWidth; }
  inline TInt VisAreaHeight() const { return iVisArea.iHeight; }

  friend TBool operator!=(const TPdfPageViewLayout& x, const TPdfPageViewLayout& y);

private:
  TSize  iVisArea; // usable size of this control

  TRect  iControlRect; 

  TBool  iShowHBar;
  TPoint iHBarPos;
  TInt   iHBarLength;
 
  TBool  iShowVBar;
  TPoint iVBarPos;
  TInt   iVBarLength;
};


/////////////////////////////////////////////////////////////////////////////
//
// MPdfPageViewCmdHandler
//
//////////////////////////////////////////////////////////////////////////////
class MPdfPageViewCmdHandler
{
public:
  virtual void CmdFirstPageL()                = 0;
  virtual void CmdPrevPageL(TBool aTop=ETrue) = 0;
  virtual void CmdNextPageL()                 = 0;
  virtual void CmdLastPageL(TBool aTop=ETrue) = 0;
  virtual void CmdCancelL()                   = 0;
  virtual void CmdTapL(TPoint aPos)           = 0;
};


/////////////////////////////////////////////////////////////////////////////
//
// TPdfPageViewMemento
//
//////////////////////////////////////////////////////////////////////////////
struct TPdfPageViewMemento: public TPdfPageControlMemento
{
  TInt iZoom;
};


/////////////////////////////////////////////////////////////////////////////
//
// CPdfPageView
//
//////////////////////////////////////////////////////////////////////////////
class CPdfPageView: public CCoeControl,
                    public MCoeControlBrushContext,
                    public MEikScrollBarObserver
{
public:
  static const TInt KZoomPage ;
  static const TInt KZoomWidth;

  void ConstructL(const TRect& aRect);
  ~CPdfPageView();

// --o CCoeControl
private:
  TInt CountComponentControls() const;
  CCoeControl* ComponentControl(TInt aIndex) const;
  void Draw(const TRect& aRect) const;
  void HandlePointerEventL(const TPointerEvent& aPointerEvent);
  TKeyResponse OfferKeyEventL(const TKeyEvent& aKeyEvent,TEventCode aType);
  void SizeChangedL();

// --o MEikScrollBarObserver
private:
  void HandleScrollEventL(CEikScrollBar* aScrollBar, TEikScrollEvent aEvent);

// --o CPdfPageView
public:
  void SetModelL(PDFDoc* aDoc);
  void SetCmdHandler(MPdfPageViewCmdHandler* aCmdHandler);

  void SetScrollBarVisibility(TInt aHBarVisibility, TInt aVBarvisibility);
  void ScrollToTop();
  void ScrollToBottom();

  void StartRenderingL(TInt aPage, const TPdfPageViewMemento& aMemento);
  void CancelRendering();

  TPoint Center() const;
  void SetCenter(TPoint aPos);
  void SetSelection(const TRect& iSelection); // user coords, in twips
  
  TInt RealZoom();

private:
  void CreateComponentControlsL();
  void DeleteComponentControls();
  void UpdateComponentControlsL();
  CEikScrollBar* CreateScrollBarL(CEikScrollBar::TOrientation aOrienation, 
                                  const TPoint& aPos, 
                                  TInt aLenght, 
                                  TEikScrollBarModel* aModel);

  void   SetScrollOffset(const TPoint& aScrollOffset);
  TPoint ScrollOffset() const;
  void   SetModelSize(TSize aSize); // reset scrollbars if the bitmap resizes
  void   SetSpanSize();  // reset scrollbars if the view resizes

  void SetDpi(TInt aPage, TInt aZoom, TInt aRotate);
  
  // translate user coords to pixel coords
  void CvtUserToDev(TReal aUserX, TReal aUserY, TPoint& aDev);
  
  // move to the previous or next page direction
  void HandlePrevNextMove(TInt aDirection);
  
  // move horizontally or vertically
  void HandleUpDownMove(TInt aDirection);
  void HandleLeftRightMove(TInt aDirection);

private:
  PDFDoc* iDoc;
  MPdfPageViewCmdHandler* iCmdHandler;

  TInt iDpi; // for current page

  // layout manager
  TPdfPageViewLayout iLayout;
  TInt iRotate; // current rotation state influences direction of movement
  
  // Bitmap Control
  CPdfPageControl* iControl;
  
  // horizontal scrollbar
  TBool iHBarVisibility;
  TEikScrollBarModel iHBarModel;
  CEikScrollBar*     iHBar;	// owns window

  // vertical scrollbar
  TBool iVBarVisibility;
  TEikScrollBarModel iVBarModel;
  CEikScrollBar*     iVBar;	 // owns window
};

#endif
