// SearchState.cpp
//
// Copyright 1999-2000 Sander van der Wal. All rights reserved.
//
// $Id: searchstate.h 1.1 2000-09-17 13:18:14+02 svdwal Exp svdwal $

#ifndef SEARCHSTATE_H_
#define SEARCHSTATE_H_

// Parents
#ifndef __E32BASE_H__
#include <e32base.h>
#endif
#include "ActivePage.h"

#ifndef __EIKDIALG_H__
#include <eikdialg.h>
#endif

// EIKON
struct SEdwinFindModel;


// engine
class PDFDoc;
class TextOutputDev;


//////////////////////////////////////////////////////////////////////////////
//
// CSearchStateObserver
//
//////////////////////////////////////////////////////////////////////////////
class MSearchStateObserver
{
public:
  virtual void NotifyFound() = 0;
  virtual void NotifyError(TInt aError) = 0;
  virtual void NotifyPage(TInt aPage) = 0;
};


//////////////////////////////////////////////////////////////////////////////
//
// CSearchState
//
//////////////////////////////////////////////////////////////////////////////
class CSearchState: public CBase,
                    public MActivePageObserver
{
public:
  enum TFindFlags { // keep in step with CEikEdwin::TFindFlags
    EFindDirectionUp  =0x01,
    EFindWholeWord    =0x02,
    EFindCaseSensitive=0x04,
    EFindAgain        =0x08,
    ENoBusyMessage    =0x10
  };
 
  CSearchState(PDFDoc* aDoc, TInt aStartPage, SEdwinFindModel* aFindModel);
  void ConstructL();
  ~CSearchState();

// --o MActivePageObserver
private:
  void NotifyPageFinishedL(TInt aError);

public:
  void SetObserver(MSearchStateObserver*);
  
  TBool FindNextL();
 
  TInt Page() const { return iPage; }
  TInt MaxPage() const;

  // current selection in user coordinates, twips
  TRect Selection() const;

private:
  // returns true if next result is fast to find
  TBool StepL();
  friend class CSearchingDialog;
  void StartSearchingL();
  void CancelSearch();

private:
  enum TState { EStart = 0,
                EStartAtSelection,
                ENextPage,
                EDisplayingNextPage,
                EPrevPage,
                EDisplayingPrevPage,
                EStopAtSelection,
                ENotFound,
                EFoundPage,
                EFound,
                EDone };
  TState iState;

  PDFDoc* iDoc;
  MSearchStateObserver* iObserver;
  SEdwinFindModel* iFindModel; 
    
  CActivePage*   iActivePage;
  TextOutputDev* iTextOut;
  
  TInt iStartPage; // first page to search
  TInt iPage;      // current page
  TInt iNumSeen;

  TBool iFound;

  // state vars
  TInt iPg;
  double iXMin1, iYMin1, iXMax1, iYMax1;

  GBool  iTop;
  TRect  iSelection; // in device coords
  TPoint iMin;
  TPoint iMax;
};


#endif
