// Fixed.h
//
// Copyright (c) 2000 Sander van der Wal
//
// $Id: Fixed.h 1.1 2000-09-17 13:35:49+02 svdwal Exp svdwal $

#ifndef FIXED_H_
#define FIXED_H_

#ifndef __E32DEF_H__
#include <e32def.h>
#endif

//
// These macros are copied straight from:
//   Application Note 33 ARM DAI033A
//   Copyright Advanced RISC Machines 1996
//   http://www.arm.com

#if 0
/* DEFINE THE MACROS */


/* The basic operations perfomed on two numbers a and b of fixed
   point q format returning the answer in q format */

#define FADD(a,b) ((a)+(b))
#define FSUB(a,b) ((a)-(b))
#define FMUL(a,b,q) (((a)*(b))>>(q))
#define FDIV(a,b,q) (((a)<<(q))/(b))


/* The basic operations where a is of fixed point q format and b is
   an integer */

#define FADDI(a,b,q) ((a)+((b)<<(q)))
#define FSUBI(a,b,q) ((a)-((b)<<(q)))
#define FMULI(a,b) ((a)*(b))
#define FDIVI(a,b) ((a)/(b))


/* convert a from q1 format to q2 format */

#define FCONV(a, q1, q2) (((q2)>(q1)) ? (a)<<((q2)-(q1)) : (a)>>((q1)-(q2)))


/* the general operation between a in q1 format and b in q2 format
   returning the result in q3 format */

#define FADDG(a,b,q1,q2,q3) (FCONV(a,q1,q3)+FCONV(b,q2,q3))
#define FSUBG(a,b,q1,q2,q3) (FCONV(a,q1,q3)-FCONV(b,q2,q3))
#define FMULG(a,b,q1,q2,q3) FCONV((a)*(b), (q1)+(q2), q3)
#define FDIVG(a,b,q1,q2,q3) (FCONV(a, q1, (q2)+(q3))/(b))


/* convert to and from floating point */

#define TOFIX(d, q) ((int)( (d)*(double)(1<<(q)) ))
#define TOFLT(a, q) ( (double)(a) / (double)(1<<(q)) )

// End of copd from Application Note 33
#endif


////////////////////////////////////////////////////
//
// TFixed: fixed point type with 16 bits mantissa
//
////////////////////////////////////////////////////

class TFixed
{
private:
  enum shift { KShift = 10 };
public:
  TFixed() {}                     // simple constructor
  TFixed(const TFixed& aVal)  // copy constructor
    : iVal(aVal.iVal) {}
  TFixed(int aVal);             // conversion from int
  TFixed(TUint aVal);           // conversion from TUint
  TFixed(double aVal);          // conversion from TReal

  TFixed& operator=(const TFixed& aX) // assignment operator
  { iVal = aX.iVal; return *this; }

  // Move internal value.
  TInt IVal() const { return iVal; }
  void SetIVal(TInt aVal) { iVal = aVal; }

  operator int();              // conversion to int
  operator unsigned int();     // to unsigned int
  operator double();           // conversion to double
  operator bool();             // conversion to bool

  TFixed& operator++();
  TFixed& operator--();

  TFixed& operator+=(TFixed aX);
  TFixed& operator-=(TFixed aX);
  TFixed& operator*=(TFixed aX);
  TFixed& operator/=(TFixed aX);

  TFixed& operator+=(TInt aX);
  TFixed& operator-=(TInt aX);
  TFixed& operator*=(TInt aX);
  TFixed& operator/=(TInt aX);

  TFixed& operator+=(TReal aX);
  TFixed& operator-=(TReal aX);
  TFixed& operator*=(TReal aX);
  TFixed& operator/=(TReal aX);

  friend TFixed operator-(TFixed aX);

  friend TFixed operator+(TFixed aX, TFixed aY);
  friend TFixed operator-(TFixed aX, TFixed aY);
  friend TFixed operator*(TFixed aX, TFixed aY);
  friend TFixed operator/(TFixed aX, TFixed aY);

  friend TFixed operator+(TFixed aX, TInt aY);
  friend TFixed operator-(TFixed aX, TInt aY);
  friend TFixed operator*(TFixed aX, TInt aY);
  friend TFixed operator/(TFixed aX, TInt aY);
  
  friend TFixed operator+(TFixed aX, TReal aY);
  friend TFixed operator-(TFixed aX, TReal aY);
  friend TFixed operator*(TFixed aX, TReal aY);
  friend TFixed operator/(TFixed aX, TReal aY);

  friend TFixed operator+(TInt aX, TFixed aY);
  friend TFixed operator-(TInt aX, TFixed aY);
  friend TFixed operator*(TInt aX, TFixed aY);
  friend TFixed operator/(TInt aX, TFixed aY);
  
  friend TFixed operator+(TReal aX, TFixed aY);
  friend TFixed operator-(TReal aX, TFixed aY);
  friend TFixed operator*(TReal aX, TFixed aY);
  friend TFixed operator/(TReal aX, TFixed aY);

  friend TBool operator==(TFixed aX, TFixed aY);
  friend TBool operator!=(TFixed aX, TFixed aY);
  friend TBool operator> (TFixed aX, TFixed aY);
  friend TBool operator>=(TFixed aX, TFixed aY);
  friend TBool operator< (TFixed aX, TFixed aY);
  friend TBool operator<=(TFixed aX, TFixed aY);

  friend TBool operator==(TFixed aX, TInt aY);
  friend TBool operator!=(TFixed aX, TInt aY);
  friend TBool operator> (TFixed aX, TInt aY);
  friend TBool operator>=(TFixed aX, TInt aY);
  friend TBool operator< (TFixed aX, TInt aY);
  friend TBool operator<=(TFixed aX, TInt aY);

  friend TBool operator==(TFixed aX, TReal aY);
  friend TBool operator!=(TFixed aX, TReal aY);
  friend TBool operator> (TFixed aX, TReal aY);
  friend TBool operator>=(TFixed aX, TReal aY);
  friend TBool operator< (TFixed aX, TReal aY);
  friend TBool operator<=(TFixed aX, TReal aY);

  friend TBool operator==(TInt aX, TFixed aY);
  friend TBool operator!=(TInt aX, TFixed aY);
  friend TBool operator> (TInt aX, TFixed aY);
  friend TBool operator>=(TInt aX, TFixed aY);
  friend TBool operator< (TInt aX, TFixed aY);
  friend TBool operator<=(TInt aX, TFixed aY);

  friend TBool operator==(TReal aX, TFixed aY);
  friend TBool operator!=(TReal aX, TFixed aY);
  friend TBool operator> (TReal aX, TFixed aY);
  friend TBool operator>=(TReal aX, TFixed aY);
  friend TBool operator< (TReal aX, TFixed aY);
  friend TBool operator<=(TReal aX, TFixed aY);

  friend TInt round(TFixed aX);

private:
  TInt iVal;
};

IMPORT_C TFixed sqrt(TFixed aX);

// inlines

inline TFixed::TFixed(int aVal) { iVal = aVal << KShift; }

inline TFixed::TFixed(TUint aVal) { iVal = aVal << KShift; }

inline TFixed::TFixed(double aVal) { iVal = int(aVal*double(1<<KShift)); }

inline TFixed::operator int() { return iVal >> KShift; }

inline TFixed::operator unsigned int() { return iVal >> KShift; }

inline TFixed::operator double() { return double(iVal) / double(1<<KShift); }

inline TFixed::operator bool() { return iVal != 0; }
 
inline TFixed& TFixed::operator++()
{
  iVal += 1 << KShift; return *this;
}

inline TFixed& TFixed::operator--()
{
  iVal -= 1 << KShift; return *this;
}

inline TFixed& TFixed::operator+=(TFixed aX) 
{ 
  iVal += aX.iVal; return *this;
}

inline TFixed& TFixed::operator-=(TFixed aX) 
{
  iVal -= aX.iVal; return *this;
}

inline TFixed& TFixed::operator*=(TFixed aX)
{
  iVal = (iVal + aX.iVal) >> KShift; return *this;
}

inline TFixed& TFixed::operator/=(TFixed aX)
{
  iVal = (iVal << KShift) / aX.iVal; return *this;
}

inline TFixed& TFixed::operator+=(TInt aX) 
{ 
  iVal += aX << KShift; return *this;
}

inline TFixed& TFixed::operator-=(TInt aX) 
{
  iVal -= aX << KShift; return *this;
}

inline TFixed& TFixed::operator*=(TInt aX)
{
  iVal *= aX; return *this;
}

inline TFixed& TFixed::operator/=(TInt aX)
{
  iVal /= aX; return *this;
}

inline TFixed& TFixed::operator+=(TReal aX) 
{ 
  iVal += int(aX*double(1<<KShift)); return *this;
}

inline TFixed& TFixed::operator-=(TReal aX) 
{
  iVal -= int(aX*double(1<<KShift)); return *this;
}

inline TFixed& TFixed::operator*=(TReal aX)
{
  iVal *= int(aX*double(1<<KShift)); return *this;
}

inline TFixed& TFixed::operator/=(TReal aX)
{
  iVal /= int(aX*double(1<<KShift)); return *this;
}


//
// Unary operators
//

inline TFixed operator-(TFixed aX)
{
  TFixed res = aX; res.iVal = -res.iVal; return res;
}

//
// Binary operators
//

inline TFixed operator+(TFixed aX, TFixed aY)
{
  TFixed res = aX; res.iVal += aY.iVal; return res;
}

inline TFixed operator-(TFixed aX, TFixed aY)
{
  TFixed res = aX; res.iVal -= aY.iVal; return res;
}

inline TFixed operator*(TFixed aX, TFixed aY)
{
  TFixed res = 0; res.iVal = (aX.iVal * aY.iVal) >> TFixed::KShift; return res;
}

inline TFixed operator/(TFixed aX, TFixed aY)
{
  TFixed res = 0; res.iVal = (aX.iVal << TFixed::KShift) / aY.iVal; return res;
}

inline TFixed operator+(TFixed aX, TInt aY){ return aX + TFixed(aY);}

inline TFixed operator-(TFixed aX, TInt aY){ return aX - TFixed(aY);}

inline TFixed operator*(TFixed aX, TInt aY)
{
  TFixed res = aX; res.iVal *= aY; return res;
}

inline TFixed operator/(TFixed aX, TInt aY)
{
  TFixed res = aX; res.iVal /= aY; return res;
}

inline TFixed operator+(TFixed aX, TReal aY){ return aX + TFixed(aY);}

inline TFixed operator-(TFixed aX, TReal aY){ return aX - TFixed(aY);}

inline TFixed operator*(TFixed aX, TReal aY){ return aX * TFixed(aY);}

inline TFixed operator/(TFixed aX, TReal aY){ return aX / TFixed(aY);}

inline TFixed operator+(TInt aX, TFixed aY) { return aY + TFixed(aX);}

inline TFixed operator-(TInt aX, TFixed aY) { return TFixed(aX) - aY;}

inline TFixed operator*(TInt aX, TFixed aY) 
{ 
  TFixed res = aY; res.iVal *= aX; return res;
}

inline TFixed operator/(TInt aX, TFixed aY)
{
  return TFixed(aX) / aY;
}

inline TFixed operator+(TReal aX, TFixed aY){ return TFixed(aX) + aY;}

inline TFixed operator-(TReal aX, TFixed aY){ return TFixed(aX) - aY;}

inline TFixed operator*(TReal aX, TFixed aY){ return TFixed(aX) * aY;}

inline TFixed operator/(TReal aX, TFixed aY){ return aX / TFixed(aY);}


inline TBool operator==(TFixed aX, TFixed aY){ return aX.iVal == aY.iVal; }

inline TBool operator!=(TFixed aX, TFixed aY){ return aX.iVal != aY.iVal; }

inline TBool operator<=(TFixed aX, TFixed aY){ return aX.iVal <= aY.iVal; }

inline TBool operator< (TFixed aX, TFixed aY){ return aX.iVal <  aY.iVal; }

inline TBool operator>=(TFixed aX, TFixed aY){ return aX.iVal >= aY.iVal; }

inline TBool operator> (TFixed aX, TFixed aY){ return aX.iVal >  aY.iVal; }

inline TBool operator==(TFixed aX, TInt aY){ return aX.iVal == (aY<<TFixed::KShift);}

inline TBool operator!=(TFixed aX, TInt aY){ return aX.iVal != (aY<<TFixed::KShift);}

inline TBool operator<=(TFixed aX, TInt aY){ return aX.iVal <= (aY<<TFixed::KShift);}

inline TBool operator< (TFixed aX, TInt aY){ return aX.iVal <  (aY<<TFixed::KShift);}

inline TBool operator>=(TFixed aX, TInt aY){ return aX.iVal >= (aY<<TFixed::KShift);}

inline TBool operator> (TFixed aX, TInt aY){ return aX.iVal >  (aY<<TFixed::KShift);}

inline TBool operator==(TFixed aX, TReal aY){ return aX == TFixed(aY); }

inline TBool operator!=(TFixed aX, TReal aY){ return aX != TFixed(aY);}

inline TBool operator<=(TFixed aX, TReal aY){ return aX <= TFixed(aY);}

inline TBool operator< (TFixed aX, TReal aY){ return aX <  TFixed(aY);}

inline TBool operator>=(TFixed aX, TReal aY){ return aX >= TFixed(aY);}

inline TBool operator> (TFixed aX, TReal aY){ return aX >  TFixed(aY);}

inline TBool operator==(TInt aX, TFixed aY){ return TFixed(aX) == aY.iVal;}

inline TBool operator!=(TInt aX, TFixed aY){ return TFixed(aX) != aY.iVal;}

inline TBool operator<=(TInt aX, TFixed aY){ return TFixed(aX) <= aY.iVal ;}

inline TBool operator< (TInt aX, TFixed aY){ return TFixed(aX) <  aY.iVal ;}

inline TBool operator>=(TInt aX, TFixed aY){ return TFixed(aX) >= aY.iVal;}

inline TBool operator> (TInt aX, TFixed aY){ return TFixed(aX) >  aY.iVal;}


inline TBool operator==(TReal aX, TFixed aY){ return TFixed(aX) == aY;}

inline TBool operator!=(TReal aX, TFixed aY){ return TFixed(aX) != aY;}

inline TBool operator<=(TReal aX, TFixed aY){ return TFixed(aX) <= aY;}

inline TBool operator< (TReal aX, TFixed aY){ return TFixed(aX) <  aY;}

inline TBool operator>=(TReal aX, TFixed aY){ return TFixed(aX) >= aY;}

inline TBool operator> (TReal aX, TFixed aY){ return TFixed(aX) >  aY;}

//
// functions
//
inline TInt round(TFixed aX) 
{ 
  TInt t = (aX.iVal < 0) ? (aX.iVal - (1 << (TFixed::KShift -1)))
                         : (aX.iVal + (1 << (TFixed::KShift -1)));
  return t >> TFixed::KShift;
}

#endif