//========================================================================
//
// GString.cc
//
// Simple variable-length string type.
//
// Copyright 1996 Derek B. Noonburg
//
//========================================================================
//
// Ported to EPOC by Sander van der Wal
//
// $Log: GString.cpp $
// Revision 1.2  2000-09-17 13:36:18+02  svdwal
// Ported
//
// Revision 1.2  2000-06-13 22:37:16+02  svdwal
// - ported
//

#ifdef __GNUC__
#pragma implementation
#endif

#ifdef __SYMBIAN32__
#  include <e32def.h>
#endif


//#include <ctype.h>
#include <string.h>

#include "GString.h"

static inline int size(int len) {
  int delta;

  delta = len < 256 ? 7 : 255;
  return ((len + 1) + delta) & ~delta;
}

inline void GString::resizeL(int aLength) {
  if (size(aLength) != size(length)) {
    s = (char*) User::ReAllocL(s, size(aLength));
  }
}


EXPORT_C void GString::ConstructL()
{
  length=0;
  s = (char*) User::AllocL(sizeof(char)*size(0));
  s[0] = '\0';
}

EXPORT_C GString* GString::NewL()
{
  GString* self = GString::NewLC();
  CleanupStack::Pop();
  return self;
}

EXPORT_C GString* GString::NewLC()
{
  GString* self = new(ELeave) GString();
  CleanupStack::PushL(self);
  self->ConstructL();
  return self;
}

EXPORT_C void GString::ConstructL(const char *aString) {
  length = User::StringLength((const TUint8*) aString);
  s = (char*) User::AllocL(sizeof(char)*size(length));
  Mem::Copy(s, aString, length + 1);
}

EXPORT_C GString* GString::NewL(const char *aString)
{
  GString* self = GString::NewLC(aString);
  CleanupStack::Pop();
  return self;
}

EXPORT_C GString* GString::NewLC(const char *aString)
{
  GString* self = new(ELeave) GString();
  CleanupStack::PushL(self);
  self->ConstructL(aString);
  return self;
}

EXPORT_C void GString::ConstructL(const char *aString, int aLength) {
  length = aLength;
  s = (char*) User::AllocL(sizeof(char)*size(length));
  Mem::Copy(s, aString, length * sizeof(char));
  s[length] = '\0';
}

EXPORT_C GString* GString::NewL(const char *aString, int aLength)
{
  GString* self = GString::NewLC(aString, aLength);
  CleanupStack::Pop();
  return self;
}

EXPORT_C GString* GString::NewLC(const char *aString, int aLength)
{
  GString* self = new(ELeave) GString();
  CleanupStack::PushL(self);
  self->ConstructL(aString, aLength);
  return self;
}

EXPORT_C void GString::ConstructL(GString *aGString) {
  length = aGString->getLength();
  s = (char*) User::AllocL(sizeof(char)*size(length));
  Mem::Copy(s, aGString->s, length+1);
}

EXPORT_C GString* GString::NewL(GString *aGString)
{
  GString* self = GString::NewLC(aGString);
  CleanupStack::Pop();
  return self;
}

EXPORT_C GString* GString::NewLC(GString *aGString)
{
  GString* self = new(ELeave) GString();
  CleanupStack::PushL(self);
  self->ConstructL(aGString);
  return self;
}

EXPORT_C void GString::ConstructL(GString *aGStr1, GString *aGStr2) {
  int n1 = aGStr1->getLength();
  int n2 = aGStr2->getLength();

  length = n1+n2;
  s = (char*) User::AllocL(sizeof(char)*size(length));
  Mem::Copy(s, aGStr1->s, n1);
  Mem::Copy(s + n1, aGStr2->s, n2 + 1);
}

EXPORT_C GString* GString::NewL(GString *aGStr1, GString *aGStr2)
{
  GString* self = GString::NewLC(aGStr1, aGStr2);
  CleanupStack::Pop();
  return self;
}

EXPORT_C GString* GString::NewLC(GString *aGStr1, GString *aGStr2)
{
  GString* self = new(ELeave) GString();
  CleanupStack::PushL(self);
  self->ConstructL(aGStr1, aGStr2);
  return self;
}

EXPORT_C GString::~GString() {
  User::Free(s); // delete[] s;
}

  // Change <i>th character.
EXPORT_C void GString::setChar(int i, char c)
{ 
  s[i] = c; 
}

EXPORT_C GString *GString::clearL() {
  s[length = 0] = '\0';
  resizeL(0);
  return this;
}

EXPORT_C GString *GString::appendL(char c) {
  resizeL(length + 1);
  s[length++] = c;
  s[length] = '\0';
  return this;
}

EXPORT_C GString *GString::appendL(GString *str) {
  int n = str->getLength();

  resizeL(length + n);
  Mem::Copy(s + length, str->getCString(), n + 1);
  length += n;
  return this;
}

EXPORT_C GString *GString::appendL(const char *str) {
  int n = User::StringLength((const TUint8*)str);

  resizeL(length + n);
  Mem::Copy(s + length, str, n + 1);
  length += n;
  return this;
}

EXPORT_C GString *GString::appendL(const char *str, int length1) {
  resizeL(length + length1);
  Mem::Copy(s + length, str, length1);
  length += length1;
  s[length] = '\0';
  return this;
}

EXPORT_C GString *GString::insertL(int i, char c) {
  
  resizeL(length + 1);
  for (register int j = length + 1; j > i; --j)
    s[j] = s[j-1];
  s[i] = c;
  ++length;
  return this;
}

EXPORT_C GString *GString::insertL(int i, GString *str) {
  int n = str->getLength();

  resizeL(length + n);
  for (register int j = length; j >= i; --j)
    s[j+n] = s[j];
  Mem::Copy(s+i, str->getCString(), n);
  length += n;
  return this;
}

EXPORT_C GString *GString::insertL(int i, char *str) {
  int n = User::StringLength((const TUint8*)str);

  resizeL(length + n);
  for (register int j = length; j >= i; --j)
    s[j+n] = s[j];
  Mem::Copy(s+i, str, n);
  length += n;
  return this;
}

EXPORT_C GString *GString::insertL(int i, char *str, int length1) {
  
  resizeL(length + length1);
  for (register int j = length; j >= i; --j)
    s[j+length1] = s[j];
  Mem::Copy(s+i, str, length1);
  length += length1;
  return this;
}

EXPORT_C GString *GString::delL(int i, int n) {
  
  if (n > 0) {
    for (register int j = i; j <= length - n; ++j)
      s[j] = s[j + n];
    resizeL(length -= n);
  }
  return this;
}

EXPORT_C GString *GString::upperCase() {

  for (register int i = 0; i < length; ++i) {
    //if (islower(s[i]))
    s[i] = User::UpperCase(s[i]);
  }
  return this;
}

EXPORT_C GString *GString::lowerCase() {
 
  for (register int i = 0; i < length; ++i) {
    //if (isupper(s[i]))
    s[i] = User::LowerCase(s[i]);
  }
  return this;
}

  // Compare two strings:  -1:<  0:=  +1:>
  // These functions assume the strings do not contain null characters.
EXPORT_C int GString::cmp(GString *str)
{ 
  return strcmp(s, str->getCString()); 
}

EXPORT_C int GString::cmpN(GString *str, int n) 
{
  return strncmp(s, str->getCString(), n); 
}

EXPORT_C int GString::cmp(char *s1)
{ 
  return strcmp(s, s1); 
}

EXPORT_C int GString::cmpN(char *s1, int n)
{ 
  return strncmp(s, s1, n); 
}
