//========================================================================
//
// GString.h
//
// Simple variable-length string type.
//
// Copyright 1996 Derek B. Noonburg
//
//========================================================================
//
// Ported to EPOC by Sander van der Wal
//
// $Id: GString.h 1.2 2000-09-17 13:36:18+02 svdwal Exp svdwal $

#ifndef GSTRING_H
#define GSTRING_H

#ifdef __GNUC__
#pragma interface
#endif

#ifdef __SYMBIAN32__
#  ifndef __E32BASE_H__
#    include <e32base.h>
#  endif
#  ifndef __S32STD_H__
#    include <s32std.h>
#  endif
#else
#  include <string.h>
#endif

class GString: public CBase {
private:
  GString(){}
public:
  // Create an empty string.
  IMPORT_C void ConstructL();
  IMPORT_C static GString* NewL();
  IMPORT_C static GString* NewLC();
  
  // Create a string from a C string.
  IMPORT_C void ConstructL(const char* aString);
  IMPORT_C static GString* NewL(const char* aString);
  IMPORT_C static GString* NewLC(const char* aString);

  // Create a string from <length1> chars at <s1>.  This string
  // can contain null characters.
  IMPORT_C void ConstructL(const char* aString, int aLength);
  IMPORT_C static GString* NewL(const char* aString, int aLength);
  IMPORT_C static GString* NewLC(const char* aString, int aLength);

  // Copy a string.
  IMPORT_C void ConstructL(GString* aGString);
  IMPORT_C static GString* NewL(GString* aGString);
  IMPORT_C static GString* NewLC(GString* aGString);

  GString *copyL() { return NewL(this); }
  GString *copyLC() { return NewLC(this); }

  // Concatenate two strings.
  IMPORT_C void ConstructL(GString *str1, GString *str2);
  IMPORT_C static GString* NewL(GString* aGStr1, GString* aStr2);
  IMPORT_C static GString* NewLC(GString* aGStr1, GString* aStr2);

  // Destructor.
  IMPORT_C ~GString();

  // Get length.
  int getLength() const { return length; }

  // Get C string.
  IMPORT_C char *getCString() { return s; }

#ifdef __SYMBIAN32__
  TPtr getTPtr() { return TPtr((TUint8 *)s,length,length); }
#endif

  // Get <i>th character.
  IMPORT_C char getChar(int i) const { return s[i]; }

  // Change <i>th character.
  IMPORT_C void setChar(int i, char c);

  // Clear string to zero length.
  IMPORT_C GString *clearL();

  // Append a character or string.
  IMPORT_C GString *appendL(char c);
  IMPORT_C GString *appendL(GString *str);
  IMPORT_C GString *appendL(const char *str);
  IMPORT_C GString *appendL(const char *str, int length1);

  // Insert a character or string.
  IMPORT_C GString *insertL(int i, char c);
  IMPORT_C GString *insertL(int i, GString *str);
  IMPORT_C GString *insertL(int i, char *str);
  IMPORT_C GString *insertL(int i, char *str, int length1);

  // Delete a character or range of characters.
  IMPORT_C GString *delL(int i, int n = 1);

  // Convert string to all-upper/all-lower case.
  IMPORT_C GString *upperCase();
  IMPORT_C GString *lowerCase();

  // Compare two strings:  -1:<  0:=  +1:>
  // These functions assume the strings do not contain null characters.
  IMPORT_C int cmp(GString *str); 
  IMPORT_C int cmpN(GString *str, int n); 
  IMPORT_C int cmp(char *s1);
  IMPORT_C int cmpN(char *s1, int n); 

private:

  int length;
  char *s;
  void resizeL(int length1);
};

#endif
