//========================================================================
//
// gfile.h
//
// Miscellaneous file and directory name manipulation.
//
// Copyright 1996 Derek B. Noonburg
//
//========================================================================
//
// Ported to EPOC by Sander van der Wal
//
// $Id: gfile.h 1.2 2000-09-17 13:36:17+02 svdwal Exp svdwal $

#ifndef GFILE_H
#define GFILE_H

#include <stdlib.h>
#include <stddef.h>
#if defined(WIN32)
#  ifdef _MSC_VER
#    include <windows.h>
#  else
#    include <kpathsea/win32lib.h>
#  endif
#elif defined(ACORN)
#elif defined(__SYMBIAN32__)
#include <dirent.h>
#else
#  include <unistd.h>
#  include <sys/types.h>
#  ifdef VMS
#    include "vms_dirent.h"
#  elif HAVE_DIRENT_H
#    include <dirent.h>
#    define NAMLEN(dirent) strlen((dirent)->d_name)
#  else
#    define dirent direct
#    define NAMLEN(dirent) (dirent)->d_namlen
#    if HAVE_SYS_NDIR_H
#      include <sys/ndir.h>
#    endif
#    if HAVE_SYS_DIR_H
#      include <sys/dir.h>
#    endif
#    if HAVE_NDIR_H
#      include <ndir.h>
#    endif
#  endif
#endif
#include "gtypes.h"

class GString;

//------------------------------------------------------------------------

// Get home directory path.
IMPORT_C GString *getHomeDir();

// Get current directory.
IMPORT_C GString *getCurrentDir();

// Append a file name to a path string.  <path> may be an empty
// string, denoting the current directory).  Returns <path>.
IMPORT_C GString *appendToPath(GString *path, char *fileName);

// Grab the path from the front of the file name.  If there is no
// directory component in <fileName>, returns an empty string.
IMPORT_C GString *grabPath(char *fileName);

// Is this an absolute path or file name?
IMPORT_C GBool isAbsolutePath(char *path);

// Make this path absolute by prepending current directory (if path is
// relative) or prepending user's directory (if path starts with '~').
IMPORT_C  GString *makePathAbsolute(GString *path);

//------------------------------------------------------------------------
// GDir and GDirEntry
//------------------------------------------------------------------------

class GDirEntry {
public:

  IMPORT_C GDirEntry(char *dirPath, char *name1, GBool doStat);
  IMPORT_C ~GDirEntry();
  GString *getName() { return name; }
  GBool isDir() { return dir; }

private:

  GString *name;		// dir/file name
  GBool dir;			// is it a directory?
};

class GDir {
public:

  IMPORT_C GDir(char *name, GBool doStat1 = gTrue);
  IMPORT_C ~GDir();
  IMPORT_C GDirEntry *getNextEntry();
  IMPORT_C void rewind();

private:

  GString *path;		// directory path
  GBool doStat;			// call stat() for each entry?
#if defined(WIN32)
  WIN32_FIND_DATA ffd;
  HANDLE hnd;
#elif defined(ACORN)
#else
  DIR *dir;			// the DIR structure from opendir()
#ifdef VMS
  GBool needParent;		// need to return an entry for [-]
#endif
#endif
};

#endif
