//========================================================================
//
// Dict.h
//
// Copyright 1996 Derek B. Noonburg
//
//========================================================================
//
// Ported to EPOC by Sander van der Wal
//
// $Id: Dict.h 1.2 2000-09-17 13:38:19+02 svdwal Exp svdwal $

#ifndef DICT_H
#define DICT_H

#ifdef __GNUC__
#pragma interface
#endif

#ifndef __E32BASE_H__
#include <e32base.h>
#endif

// --o PdfLib
#include "Object.h"

//------------------------------------------------------------------------
// Dict
//------------------------------------------------------------------------
struct DictEntry {
  char *key;
  Object val;
};

class Dict: public CBase {
public:

  // Constructor.
  Dict();

  // Destructor.
  ~Dict();

  // Reference counting.
  int incRef() { return ++ref; }
  int decRef() { return --ref; }

  // Get number of entries.
  int getLength() const { return length; }

  // Add an entry.  NB: does not copy key.
  void addL(char *key, Object *val);

  // Check if dictionary is of specified type.
  GBool is(char *type);

  // Look up an entry and return the value.  Returns a null object
  // if <key> is not in the dictionary.
  Object *lookupL(char *key, Object *obj);
  Object *lookupNFL(char *key, Object *obj);

  // Iterative accessors.
  char *getKey(int i);
  Object *getValL(int i, Object *obj);
  Object *getValNFL(int i, Object *obj);

private:

  DictEntry *entries;		// array of entries
  int size;			// size of <entries> array
  int length;			// number of entries in dictionary
  int ref;			// reference count

  DictEntry *find(char *key);
};

#endif
