// EdgeTable.h
//
// Original code by Michael Abrash, Dr. Dobbs Journal #177, June 1991
//
// Ported to EPOC by Sander van der Wal,  12 june 2000
// Adapted to add rectangles to a region.
//
// $Id: edgetable.h 1.1 2000-09-17 13:20:25+02 svdwal Exp svdwal $

#ifndef EDGETABLE_H_
#define EDGETABLE_H_

// parent
#ifndef __E32BASE_H__
#include <e32base.h>
#endif

// interface
#ifndef __GDI_H__
#include <gdi.h>
#endif

// forward declarations

class CPolygon;

struct TEdgeState;  // edge information

///////////////////////////////////
//
// CEdgeState
//
///////////////////////////////////
class CEdgeTable: public CBase
{
public:
  CEdgeTable();
  void ConstructL(const CArrayFixFlat<TPoint>& aPolygon);
  ~CEdgeTable();

  TInt StartY() const;
  TBool Empty() const { return !(iGETPtr != 0 || iAETPtr != 0); }

  void AddScanlineToRegion(TInt aYToScan, RRegion& aRegion, CGraphicsContext::TFillRule aFillRule);
  void NextScanline();

private:
  void MoveXSortedToAET(TInt aYToMove);
  void RectOutAET(TInt aYToScan, RRegion& aRegion, CGraphicsContext::TFillRule aFillRule);
  void AdvanceAET();
  void XSortAET();

private:
  TEdgeState* iTable; // all edges, either in the GET or the AET

  TEdgeState* iGETPtr; // Global Edge Table
  TEdgeState* iAETPtr; // Active Edge Table
};

#endif
