// FbsBitmapOutputDev.cpp
//
// Copyright 1999  Sander van der Wal
//
// $Id: fbsbitmapoutputdev.cpp 1.3 2000-09-24 21:16:40+02 svdwal Exp svdwal $
//
// $Log: fbsbitmapoutputdev.cpp $
// Revision 1.3  2000-09-24 21:16:40+02  svdwal
// Bug fixes related to clipping region
//
// Revision 1.2  2000-09-21 13:52:04+02  svdwal
// Device created and destroyed at start and end of page.
// Output state also saves both botmap graphics contexts
//
// Revision 1.1  2000-09-17 13:20:26+02  svdwal
// Initial checkin
//

#ifdef __GNUC__
#pragma implementation
#endif

// Interfaces
#include "FbsBitmapOutputDev.h"
#include "FbsBitmapOutputDevObserver.h"
#include "FbsBitmapMemento.h"

// BASE
#ifndef __E32MATH_H__
#include <e32math.h>
#endif
 
// CONE
#ifndef __COEMAIN_H__
#include <coemain.h>
#endif

// EIKON
#ifndef __EIKENV_H__
#include <eikenv.h>
#endif
#ifndef __EIKMSG_H__
#include <eikmsg.h>
#endif

// View
#include "GDIView.h"


// PDF
#include "Error.h"
#include "GfxState.h"
#include "GfxFont.h"
#include "FontEncoding.h"
#include "PDFGLobal.h"
#include "Pdf.rsg"
#include "Pdf.hrh"

// PROFILING
#include "PROFILE.H"


// Local
static inline TInt round(double x) { return (TInt) (x+0.5); }
static const TInt KBusyInitialDelay = 100000; // 0.1 second


//------------------------------------------------------
// CBitmapTextView
//
// Implements CRotatedTextView for a CFbsBitmapOutputDev
//------------------------------------------------------
class CBitmapTextView: public CRotatedTextView
{
public:
  static CGDITextView* NewL(const CGDITextView& aOriginal, TInt aRotate, CGraphicsDevice* aRotDev, TGDIFontNameSubst* aFontNameSubst, CWsBitmap* aRotMap);

private:
  CBitmapTextView(const CBitmapTextView& aOriginal);
  CBitmapTextView(const CGDITextView& aOriginal, TInt aRotate, CGraphicsDevice* aRotDev, TGDIFontNameSubst* aFontNameSubst, CWsBitmap* aRotMap);

// --o CRotatedTextView
public:
  CGDITextView* CopyL();

private:
  void DrawRotatedStringL(TBitmapUtil* aUtil0, TPoint aZeroEnd);

private:
  // the original rotated bitmap
  CWsBitmap* iRotMap;
};


CBitmapTextView::CBitmapTextView(const CBitmapTextView& aOriginal)
  : CRotatedTextView(aOriginal)
{}


CBitmapTextView::CBitmapTextView(const CGDITextView& aOriginal, TInt aRotate, CGraphicsDevice* aRotDev, TGDIFontNameSubst* aFontNameSubst, CWsBitmap* aRotMap)
  : CRotatedTextView(aOriginal, aRotate, aRotDev, aFontNameSubst),
    iRotMap(aRotMap)
{}


CGDITextView* CBitmapTextView::NewL(const CGDITextView& aOriginal, TInt aRotate, CGraphicsDevice* aRotDev, TGDIFontNameSubst* aFontNameSubst, CWsBitmap* aRotMap)
{
  if (aRotate == 0)
    return new(ELeave) CGDITextView(aOriginal);
  else {
    CBitmapTextView* self = new(ELeave) CBitmapTextView(aOriginal, aRotate, aRotDev, aFontNameSubst, aRotMap);
    CleanupStack::PushL(self);
    self->ConstructL();
    CleanupStack::Pop(); // self
    return self;
  }
}


CGDITextView* CBitmapTextView::CopyL()
{
  return CBitmapTextView::NewL(*this, this->iRotate, this->iRotDev, this->iFontNameSubst, this->iRotMap);
}


void CBitmapTextView::DrawRotatedStringL(TBitmapUtil* aUtil0, TPoint aZeroEnd)
{
  TBitmapUtil* util = new(ELeave) TBitmapUtil(iRotMap);
 
  register TInt y0;
  register TInt pixel = FindTextPixel(iRotMap->DisplayMode());

  switch (iRotate) {
  case 90:
    iRotStart.iX += iFont->AscentInPixels(); 
    util->Begin(iRotStart, *aUtil0);
    for (y0 = 0; y0 < aZeroEnd.iY - iZeroStart.iY; y0++) {
      aUtil0->SetPos(TPoint(iZeroStart.iX, iZeroStart.iY + y0));
      util->SetPos(TPoint(iRotStart.iX-y0, iRotStart.iY));
      for (register TInt x0 = iZeroStart.iX; x0 < aZeroEnd.iX; x0++) {
        if (aUtil0->GetPixel() == 0)
          util->SetPixel(pixel);
        util->IncYPos();
        aUtil0->IncXPos();
      }
    }
    break;
  
  case 180:
    iRotStart.iY += iFont->AscentInPixels();
    util->Begin(iRotStart, *aUtil0);
    for (y0 = 0; y0 < aZeroEnd.iY - iZeroStart.iY; y0++) {
      aUtil0->SetPos(TPoint(iZeroStart.iX, iZeroStart.iY + y0));
      util->SetPos(TPoint(iRotStart.iX, iRotStart.iY - y0));
      for (register TInt x0 = iZeroStart.iX; x0 < aZeroEnd.iX; x0++) {
        if (aUtil0->GetPixel() == 0)
          util->SetPixel(pixel);
        util->DecXPos();
        aUtil0->IncXPos();
      }
    }
    break;

  case 270:
    iRotStart.iX -= iFont->AscentInPixels();
    util->Begin(iRotStart, *aUtil0);
    for (y0 = 0; y0 < aZeroEnd.iY - iZeroStart.iY; y0++) {
      aUtil0->SetPos(TPoint(iZeroStart.iX, iZeroStart.iY + y0));
      util->SetPos(TPoint(iRotStart.iX + y0, iRotStart.iY));
      for (register TInt x0 = iZeroStart.iX; x0 < aZeroEnd.iX; x0++) {
        if (aUtil0->GetPixel() == 0)
          util->SetPixel(pixel);
        util->DecYPos();
        aUtil0->IncXPos();
      }
    }
    break;
  
  default:
    __ASSERT_DEBUG(EFalse, User::Invariant());
    break;
  }

  util->End();

  delete util;
}

//
// CFbsBitmapOutputState
//
TFbsBitmapOutputState::TFbsBitmapOutputState(CFbsBitmapOutputDev* aDev)
  :TGDIOutputState(aDev),
   iBitTextGc(aDev->iBitTextGc),
   iBitGraphicsGc(aDev->iBitGraphicsGc)
{}


//-------------------------------------
//
// CFbsBitmapOutputDev
//
//-------------------------------------
CFbsBitmapOutputDev::CFbsBitmapOutputDev(MFbsBitmapOutputDevObserver* aObserver)
  : CGDIOutputDev(ETrue, ETrue, ETrue),
   iObserver(aObserver)
{}


void CFbsBitmapOutputDev::ConstructL(CWsBitmap* aBitmap, TDisplayMode aDisplayMode)
{ 
  iBitmap      = aBitmap;
  iDisplayMode = aDisplayMode;

  iBusyMsg = CCoeEnv::Static()->AllocReadResourceL(R_IMAGE);
}


CFbsBitmapOutputDev::~CFbsBitmapOutputDev()
{
  delete iBusyMsg;
  delete iBitDev;
}


//
// OutputDev reimplementation
//

//----- initialization and control

// Start a page.
void CFbsBitmapOutputDev::startPageL(int /*pageNum*/, GfxState *state)
{
  // calculate size in pixels of bitmap
  TSize size(round(state->getPageWidth()), round(state->getPageHeight()));

  // Set the size of the bitmap
  if (size != iBitmap->SizeInPixels()) {
    User::LeaveIfError(iBitmap->Resize(size));
    iBitmap->SetSizeInTwips(iPageSizeInTwips);
  }

  // recreate the bitmap device
  iBitDev = CFbsBitmapDevice::NewL(iBitmap);
  iDev = iBitDev;
  CheckTypefacesL();

  // set the clipping path to the bitmap
  iClipRegion.AddRect(TRect(TPoint(0,0), size));
  if (iClipRegion.CheckError()) {
    // because we didn't use ForceError() the error must be a memory shortage
    User::LeaveNoMemory();
  }
 
  // create the text and the graphic context

  User::LeaveIfError(iBitDev->CreateContext(iBitTextGc));  
  iTextGc = iBitTextGc;
  iBitTextGc->SetBrushStyle(CGraphicsContext::ESolidBrush);
  iBitTextGc->SetBrushColor(KRgbWhite);
  iBitTextGc->Clear();
  iBitTextGc->SetBrushStyle(CGraphicsContext::ENullBrush);
  if (!iRenderText) { 
    delete iBitTextGc;
    iBitTextGc = 0;
    iTextGc = 0;
  }

  if (iRenderLineDrawings || iRenderImages) {
    User::LeaveIfError(iBitDev->CreateContext(iBitGraphicsGc));
    iGraphicsGc = iBitGraphicsGc;
    iBitGraphicsGc->SetBrushStyle(CGraphicsContext::ESolidBrush);
  }

  InitPageL();

  if (iObserver)
    iObserver->NotifyDump();

  if (!iClip)
    return;
    
  if (iRenderText)
    iBitTextGc->SetClippingRegion(&iClipRegion);

  if ((iRenderLineDrawings || iRenderImages))
    iBitGraphicsGc->SetClippingRegion(&iClipRegion);
}

void CFbsBitmapOutputDev::endPage()
{
  CGDIOutputDev::endPage();
  delete iBitDev;
  iBitDev = 0;
  iDev = 0;
}

// Dump page contents to display.
void CFbsBitmapOutputDev::dump()
{
  if (iObserver)
    iObserver->NotifyDump();
}


//----- link borders
void CFbsBitmapOutputDev::drawLinkBorder(double x1, double y1, 
                                         double x2, double y2,
                                         double w)
{
  // create a device and a context
  CFbsBitmapDevice* gd = CFbsBitmapDevice::NewL(iBitmap);
  CleanupStack::PushL(gd);
  
  CFbsBitGc* gc;
  User::LeaveIfError(gd->CreateContext(gc));
  
  // 1 is default
  if (w > 1.0)
    gc->SetPenSize(TSize(round(w),round(w)));
  gc->SetPenColor(KRgbDarkBlue);
  
  TRect r;
  
  cvtUserToDev(x1, y1, &r.iTl.iX, &r.iTl.iY);
  cvtUserToDev(x2, y2, &r.iBr.iX, &r.iBr.iY);
  r.Normalize();
  gc->DrawRect(r);
  
  delete gc;
  
  CleanupStack::PopAndDestroy(); // delete gd;
}


//----- save/restore graphics state
void CFbsBitmapOutputDev::saveState(GfxState* /*state*/)
{
  ClearPath();

  CFbsBitGc* textGc = 0;
  CGDITextView* textView = 0;
  if (iRenderText) {
    User::LeaveIfError(iBitDev->CreateContext(textGc));
    CleanupStack::PushL(textGc);
    textGc->CopySettings(*iBitTextGc);
    
    textView = iTextView->CopyL();
    CleanupStack::PushL(textView);
  }

  CFbsBitGc* graphicsGc = 0;
  if (iRenderLineDrawings || iRenderImages) {
    User::LeaveIfError(iBitDev->CreateContext(graphicsGc));
    CleanupStack::PushL(graphicsGc);
    graphicsGc->CopySettings(*iBitGraphicsGc);
  }

  // save current state, push on the state stack
  TFbsBitmapOutputState *s = new(ELeave) TFbsBitmapOutputState(this);

  // set the new grpahic contexts
  if (iRenderLineDrawings || iRenderImages) {
    CleanupStack::Pop(); // graphicsGc
    iBitGraphicsGc = graphicsGc;
    iGraphicsGc = graphicsGc;
  }

  if (iRenderText) {
    CleanupStack::Pop(2); // textView, textGc
    iBitTextGc = textGc;
    iTextGc = textGc;
    iTextView = textView;
  }

  iClipRegion = RPath();
  if (iClip) {
    
    // new clippath
    iClipRegion.Union(s->iClipRegion);
    if (iClipRegion.CheckError()) {
      // because we didn't use ForceError() the error must be a memory shortage
      //      iClipRegion.Clear();
      User::LeaveNoMemory();
    }

    if (iRenderLineDrawings || iRenderImages)
      iBitGraphicsGc->SetClippingRegion(&iClipRegion);

    if (iRenderText)
     iBitTextGc->SetClippingRegion(&iClipRegion);
  }
}


void CFbsBitmapOutputDev::restoreState(GfxState* /*state*/)
{
  if (iSave) {
    // it is known that we really pushed a TFbsBitmapOutputState;
    TFbsBitmapOutputState* s = (TFbsBitmapOutputState*) iSave;
    // restore state, pop stack
    iBitGraphicsGc = s->iBitGraphicsGc;
    iBitTextGc     = s->iBitTextGc;
    iSave->RestoreAndPop(this);
    if (iClip) {

      if (iRenderLineDrawings || iRenderImages)
        iBitGraphicsGc->SetClippingRegion(&iClipRegion);

      if (iRenderText)
       iBitTextGc->SetClippingRegion(&iClipRegion);
    }
  }
}


//----- path painting
void CFbsBitmapOutputDev::stroke(GfxState *state)
{
  if (!iRenderLineDrawings && !iClip) 
    return;

  ConvertPathL(state);

  if (!iRenderLineDrawings)
    return;

  if (iClipPath)
    MakeClipRegion();

  // Draw subpaths.
  for (register TInt i = 0; i<iPath->Count(); i++) {
    CPolygon* polygon = iPath->At(i);
    switch (polygon->Count()) {
    case 0:
    case 1:
      break;

    case 2: {
      // this is a fix for a very strange error: if the path is just one or two pixels
      // long, a filled circle is drawn instead.
      TPoint p0 = polygon->At(0);
      TPoint p1 = polygon->At(1);
      if (Abs(p0.iX-p1.iX)>2 || Abs(p0.iY-p1.iY)>2)
        iGraphicsGc->DrawPolyLine(polygon);
    }
    break;
      
    default:
      iGraphicsGc->DrawPolyLine(polygon);
      break;
    
    }
  }
}


void CFbsBitmapOutputDev::FillPathL(CGraphicsContext::TFillRule aFillRule)
{  
  StartBusyMsgWinLC();
  
  // Fill subpaths.
  for (register TInt i = 0; i<iPath->Count(); i++) 
    iGraphicsGc->DrawPolygon(iPath->At(i), aFillRule);

  StopBusyMsgWin();
}


//----- path clipping
void CFbsBitmapOutputDev::SetTextClipPath()
{
  iBitTextGc->SetClippingRegion(&iClipRegion);
}


void CFbsBitmapOutputDev::SetGraphicsClipPath()
{
  iBitGraphicsGc->SetClippingRegion(&iClipRegion);
}


//----- text drawing
void CFbsBitmapOutputDev::beginString(GfxState*state, GString* /* s */)
{
  if (iRenderText)
    iTextView->BeginString(state);
}


void CFbsBitmapOutputDev::endString(GfxState* state)
{
  if (iRenderText)
    iTextView->EndString(state);
}


void CFbsBitmapOutputDev::drawImageMask(GfxState *state, Stream *str, int width,
		                        int height, GBool invert, GBool inlineImg)      
{
  // get image position and size in device coordinates
  TRect rect;
  TBool rotate, xFlip, yFlip;
  GetImagePlacementInfo(state, rect, rotate, xFlip, yFlip);

  // check for tiny images and throw them away
  if (!iRenderImages || rect.Width() == 0 || rect.Height() == 0) {
    OutputDev::drawImageMask(state, str, width, height, invert, inlineImg);
    return;
  }
   
  if (iOutlineImages) {
  
    iGraphicsGc->SetPenStyle(CGraphicsContext::EDottedPen);
    iGraphicsGc->SetPenColor(KRgbBlack);
    
    iGraphicsGc->SetBrushStyle(CGraphicsContext::ESquareCrossHatchBrush);
    iGraphicsGc->SetBrushColor(KRgbGray);

    if (iClipPath) {
      for (TInt i = 0; i < iClipPath->Count(); i++)
        iGraphicsGc->DrawPolygon(iClipPath->At(i), iFillRule);
    }
    else
      iGraphicsGc->DrawRect(rect);
    
    iGraphicsGc->SetPenStyle(iLineDash);
    iGraphicsGc->SetPenColor(iStrokeColor);
    
    iGraphicsGc->SetBrushStyle(CGraphicsContext::ESolidBrush);
    iGraphicsGc->SetBrushColor(iFillColor);
  
  }
  else {
    
    // start displaying the image message
    StartBusyMsgWinLC();

    if (iClipPath)
      MakeClipRegion();

    // draw directly into the the bitmap
    DoDrawImageMaskL(str, TSize(width, height), invert,
                     rotate, xFlip, yFlip, 
                     rect, iBitmap, iDev->DisplayMode());
   
    StopBusyMsgWin();
  }
}


void CFbsBitmapOutputDev::drawImage(GfxState *state, Stream *str, int width,
                              int height, GfxImageColorMap *colorMap,
                              GBool inlineImg)
{
  // get image position and size in device coordinates
  TRect rect;
  TBool rotate, xFlip, yFlip;
  GetImagePlacementInfo(state, rect, rotate, xFlip, yFlip);

  // check for tiny images and throw them away
  if (!iRenderImages || rect.Width() == 0 || rect.Height() == 0) {
    OutputDev::drawImage(state, str, width, height, colorMap, inlineImg);
    return;
  }
  
  if (iOutlineImages) {

    iGraphicsGc->SetPenStyle(CGraphicsContext::ESolidPen);
    iGraphicsGc->SetPenColor(KRgbBlack);
    iGraphicsGc->SetBrushColor(KRgbDarkGray);

    if (iClipPath) {
      for (TInt i = 0; i < iClipPath->Count(); i++)
        iGraphicsGc->DrawPolygon(iClipPath->At(i), iFillRule);
    }
    else
      iGraphicsGc->DrawRect(rect);

    iGraphicsGc->SetPenStyle(iLineDash);
    iGraphicsGc->SetPenColor(iStrokeColor);
    iGraphicsGc->SetBrushColor(iFillColor);
  }
  else {
   
    // image drawing takes a long time
    StartBusyMsgWinLC();
  
    if (!iClip || (!iClipPath && iClipRegion.IsRect())) {
      // draw scaled image directly into the bitmap
      TRect clipRect(rect);
      clipRect.Intersection(TRect(TPoint(0,0), iDev->SizeInPixels()));
      if (!iClipPath && iClipRegion.IsRect())
        clipRect.Intersection(iClipRegion.Rect());
      DoDrawImageL(str, TSize(width, height), colorMap,
                   rotate, xFlip, yFlip, iDitherImages, rect, iBitmap, iDev->DisplayMode(),
                   clipRect);
    }
    else {
      // draw scaled image in separate bitmap, then clip it 
    
      // create the bitmap, in the color mode of the graphics device
      // and the size in device coordinates
      TDisplayMode dstDisplayMode = iDev->DisplayMode();
      TRect dstRect(TSize(rect.Width(), rect.Height()));
      CWsBitmap* dstBitmap = new(ELeave) CWsBitmap(CCoeEnv::Static()->WsSession());
      CleanupStack::PushL(dstBitmap);
      User::LeaveIfError(dstBitmap->Create(dstRect.iBr.AsSize(), dstDisplayMode));  
      dstBitmap->SetSizeInTwips(iDev);
    
      DoDrawImageL(str, TSize(width, height), colorMap,
                   rotate, xFlip, yFlip, iDitherImages,
                   dstRect, dstBitmap, dstDisplayMode, dstRect);
  
      if (iClipPath) {
        // draw bitmap as brush of Polygon
        iGraphicsGc->SetBrushOrigin(rect.iTl);
        iGraphicsGc->UseBrushPattern(dstBitmap);
        iGraphicsGc->SetBrushStyle(CGraphicsContext::EPatternedBrush);
        iGraphicsGc->SetPenStyle(CGraphicsContext::ENullPen);
        for (TInt i = 0; i < iClipPath->Count(); i++) { 
          iGraphicsGc->DrawPolygon(iClipPath->At(i), iFillRule);
        }
        iGraphicsGc->DiscardBrushPattern();
        iGraphicsGc->SetBrushStyle(CGraphicsContext::ESolidBrush);
        iGraphicsGc->SetPenStyle(iLineDash);
      }
      else
         // draw bitmap
        iBitGraphicsGc->BitBlt(rect.iTl, dstBitmap, dstRect);

      CleanupStack::PopAndDestroy(); // dstBitmap
    }
    
    StopBusyMsgWin();

  } // if (iOutlineImages)
}


void CFbsBitmapOutputDev::DoDrawImageMaskL(Stream *aStr, TSize aImgSize, TBool aInvert, 
                                           TBool aRotate, TBool aXFlip, TBool aYFlip,
                                           const TRect& aDevRect, CWsBitmap* aDstBitmap,
                                           TDisplayMode aDisplayMode)
{
  // initialize the image stream
  ImageStream* imgStr = new(ELeave) ImageStream(aStr, aImgSize.iWidth, 1, 1);
  CleanupStack::PushL(imgStr);
  imgStr->ConstructL();
  imgStr->reset();

  // the size of the unrotated image in device measurements
  TSize devSize(aRotate ? aDevRect.Height(): aDevRect.Width(),
                aRotate ? aDevRect.Width() : aDevRect.Height());

  // Bresenham parameters
  TInt px1 = devSize.iWidth / aImgSize.iWidth;
  TInt px2 = devSize.iWidth - px1 * aImgSize.iWidth;
  TInt py1 = devSize.iHeight / aImgSize.iHeight;
  TInt py2 = devSize.iHeight - py1 * aImgSize.iHeight;

  // It is possible for images to be partly outside the destination bitmap.
  // The bitmaputils will wrap such an image.
  // We assume here there's always some intersection.
  TRect clipRect(aDevRect);
  clipRect.Intersection(TRect(TPoint(0,0), aDstBitmap->SizeInPixels()));

  TInt pixel = iTextView->FindTextPixel(aDisplayMode);
  
  // position in destination bitmap
  TPoint pos;

  // pointer to member function IncXPos, DecXPos, IncYPos or DecYPos
  typedef void (TBitmapUtil::*TBitmapUtilNextPos)();
  TBitmapUtilNextPos nextXPos;
  TBitmapUtilNextPos nextYPos;
  TBitmapUtilNextPos prevYPos;

  TBitmapUtil* util = new(ELeave) TBitmapUtil(aDstBitmap);
   
  // first line (column)
  TInt x = aXFlip ? devSize.iWidth -1: 0;
  TInt y = aYFlip ? devSize.iHeight-1: 0;
  if (aRotate) {
    pos.iY  = aDevRect.iTl.iY + x;
    pos.iX  = aDevRect.iTl.iX + y;
    nextXPos= aXFlip ? &TBitmapUtil::DecYPos : &TBitmapUtil::IncYPos;
    nextYPos= aYFlip ? &TBitmapUtil::DecXPos : &TBitmapUtil::IncXPos;
    prevYPos= aYFlip ? &TBitmapUtil::IncXPos : &TBitmapUtil::DecXPos;
  }
  else {
    pos.iX   = aDevRect.iTl.iX + x;
    pos.iY   = aDevRect.iTl.iY + y;
    nextXPos = aXFlip ? &TBitmapUtil::DecXPos : &TBitmapUtil::IncXPos;
    nextYPos = aYFlip ? &TBitmapUtil::DecYPos : &TBitmapUtil::IncYPos;
    prevYPos = aYFlip ? &TBitmapUtil::IncYPos : &TBitmapUtil::DecYPos;
  }
  util->Begin(pos);
  TInt qy = 0;

  for (TInt i=0; i < aImgSize.iHeight; i++) {

    // vertical (horizontal) Bresenham
    TInt dy = py1;
    if ((qy += py2) >= aImgSize.iHeight) {
      ++dy;
      qy -= aImgSize.iHeight;
    }
     
    TBool clip = (dy == 0);
    if (!clip) {
    
      // first column (line)
      if (aRotate) {
        pos.iX  = aDevRect.iTl.iX + y;
        clip = pos.iX < clipRect.iTl.iX || clipRect.iBr.iX <= pos.iX;
      }
      else {
        pos.iY  = aDevRect.iTl.iY + y;
        clip = pos.iY < clipRect.iTl.iY || clipRect.iBr.iY <= pos.iY;
      }
    }

    if (clip) {
      // drop a line (column)  
      imgStr->skipLine();
    } 
    else {
      
      // draw a line (column)
      x = aXFlip ? devSize.iWidth - 1 : 0;
      if (aRotate) {
        pos.iY  = aDevRect.iTl.iY + x;
        nextYPos= aYFlip ? &TBitmapUtil::DecXPos : &TBitmapUtil::IncXPos;
        prevYPos= aYFlip ? &TBitmapUtil::IncXPos : &TBitmapUtil::DecXPos;
      }
      else {
        pos.iX  = aDevRect.iTl.iX + x;
        nextYPos = aYFlip ? &TBitmapUtil::DecYPos : &TBitmapUtil::IncYPos;
        prevYPos = aYFlip ? &TBitmapUtil::IncYPos : &TBitmapUtil::DecYPos;
      }
      util->SetPos(pos);
      TInt qx = 0;

      // read a line (column)
      for (register TInt j= 0; j< aImgSize.iWidth; j++) {    

        // horizontal (vertical) Bresenham
	TInt dx = px1;
	if ((qx += px2) >= aImgSize.iWidth) {
	  ++dx;
	  qx -= aImgSize.iWidth;
	}

        Guchar pixBuf;
        imgStr->getPixel(&pixBuf);
        if (dx > 0) {
        
          // draw image pixel
          if (aInvert)
            pixBuf ^= 1;  

          clip = pixBuf != 0;
          if (!clip) {
            if (aRotate)
              clip = aDevRect.iTl.iY + x < clipRect.iTl.iY || clipRect.iBr.iY <= aDevRect.iTl.iY + x;
            else
              clip = aDevRect.iTl.iX + x < clipRect.iTl.iX || clipRect.iBr.iX <= aDevRect.iTl.iX + x;
          }
          if (dy == 1 && dx == 1) {
            if (!clip)
              util->SetPixel(pixel);
            (util->*nextXPos)();
          }
          else {
            for (register TInt ix = 0; ix < dx; ix++) {
              for (register TInt iy = 0; iy < dy; iy++) {
                if (!clip)
                  util->SetPixel(pixel);
                (util->*nextYPos)();
              }
              if (!clip)
                util->SetPixel(pixel);
              (util->*nextXPos)();
              // swap direction of y
              TBitmapUtilNextPos tmp = nextYPos;
              nextYPos = prevYPos;
              prevYPos = tmp;
            }
          }
          x += aXFlip ? -dx:dx;
         
        } // if (dx > 0)
      }
      // next line (column)
      y += aYFlip ? -dy: dy;
    }
  } // for (i=0; ...
  
  util->End();
  delete util;
  
  CleanupStack::PopAndDestroy(); //  imgStream
}


void CFbsBitmapOutputDev::CreateTextViewL(TInt aRotate)
{ 
  CGDITextView* oldView = iTextView;

  iTextView = CBitmapTextView::NewL(*oldView, aRotate, iBitDev, iFontNameSubst, iBitmap);

  delete oldView;
}


void CFbsBitmapOutputDev::InitPageRendering(TSize aSizeInTwips, const TFbsBitmapMemento& aMemento)
{
  if (iPageSizeInTwips != aSizeInTwips) {
    iPageSizeInTwips = aSizeInTwips;
    iBitmap->SetSizeInTwips(iPageSizeInTwips);
  }
 
  iRenderText = aMemento.iRenderText;
  iRenderLineDrawings = aMemento.iRenderLineDrawings;
  iRenderImages  = aMemento.iRenderImages >= 1;
  iUseDrawChar   = aMemento.iDrawAccurate;
  iOutlineImages = aMemento.iRenderImages >= 2;
  iDitherImages  = aMemento.iDitherImages;
  iClip          = aMemento.iClip;
}


void CFbsBitmapOutputDev::StartBusyMsgWinLC()
{
  CEikonEnv* env = CEikonEnv::Static();
  if (!iBusyMsg)
    iBusyMsg = env->AllocReadResourceL(R_IMAGE);
  
  CEikBusyMsgWin* win = new(ELeave) CEikBusyMsgWin(*env);
  CleanupStack::PushL(win);
  win->ConstructL(env->RootWin());
  win->StartDisplaySpecifyingInitialDelay(*iBusyMsg, EHRightVBottom, KBusyInitialDelay);
}


void CFbsBitmapOutputDev::StopBusyMsgWin()
{
  CleanupStack::PopAndDestroy();
}