// FbsBitmapOutputDev.h
//
// Copyright 1999 Sander van der Wal
//
// $Id: fbsbitmapoutputdev.h 1.2 2000-09-21 13:52:05+02 svdwal Exp svdwal $

#ifndef FBSBITMAPOUTPUTDEV_H
#define FBSBITMAPOUTPUTDEV_H

#ifdef __GNUC__
#pragma interface
#endif

// Parents
#include "GDIOutputDev.h"


class CWsBitmap;
class CFbsBitmapDevice;
class CFbsBitGc;


// separate interfaces
class MFbsBitmapOutputDevObserver;
struct TFbsBitmapMemento;

//
// TFbsBitmapOutputState
//
class CFbsBitmapOutputDev;
class TFbsBitmapOutputState: public TGDIOutputState
{
public:
  TFbsBitmapOutputState(CFbsBitmapOutputDev* aDev);

public:
  CFbsBitGc* iBitGraphicsGc;
  CFbsBitGc* iBitTextGc;
};


//------------------------------------------------------------------------
// CFbsBitmapOutputDev
//------------------------------------------------------------------------
class CFbsBitmapOutputDev: public CGDIOutputDev
{
public:
  CFbsBitmapOutputDev(MFbsBitmapOutputDevObserver* aObserver);
  void ConstructL(CWsBitmap* aBitmap, TDisplayMode aDisplayMode);
  ~CFbsBitmapOutputDev();

// --o OutputDev

  //----- get info about output device

  // This device uses drawString()
  GBool useDrawChar() { return iUseDrawChar; };

  // Start a page.
  void startPageL(int pageNum, GfxState *state);
  void endPage();
  
  // Dump page contents to display.
  void dump();
  
  void drawLinkBorder(double x1, double y1, double x2, double y2, double w);
  
  //----- save/restore graphics state
  void saveState(GfxState *state);
  void restoreState(GfxState *state);

  //----- path painting
  void stroke(GfxState *state);
 
  //----- text drawing
  void beginString(GfxState *state, GString *s);
  void endString(GfxState *state);

  //----- image drawing
  void drawImageMask(GfxState *state, Stream *str,
                     int width, int height, GBool invert,
		     GBool inlineImg);
  void drawImage(GfxState *state, Stream *str, 
                 int width, int height, GfxImageColorMap *colorMap,
                 GBool inlineImg);

// --o CGDIOutputDev
private:
  void CreateTextViewL(TInt aRotate);

  void FillPathL(CGraphicsContext::TFillRule aFillRule);
  void SetTextClipPath();
  void SetGraphicsClipPath();

// --o CFbsBitmapOutputDev
public:
  void InitPageRendering(TSize aSizeInTwips, const TFbsBitmapMemento& aMemento);
  void SetUseDrawChar(TBool aUseDrawChar);

private:
  void StartBusyMsgWinLC();
  void StopBusyMsgWin();
 
  void DoDrawImageMaskL(Stream *aStr, TSize aImgSize, TBool aInvert,
                        TBool aRotate, TBool aXFlip, TBool aYFlip,
                        const TRect& aDevRect, CWsBitmap* aDstBitmap,
                        TDisplayMode aDisplayMode);

private:
  friend class TFbsBitmapOutputState;

  MFbsBitmapOutputDevObserver* iObserver;
  
  GBool iUseDrawChar;   // user can choose the speed
  TBool iOutlineImages;
  TBool iDitherImages;

  TDesC* iBusyMsg; // to show we are too busy responding to events

  // the page is drawn on a bitmap
  CWsBitmap*        iBitmap;
  CFbsBitmapDevice* iBitDev;
  CFbsBitGc*        iBitGraphicsGc;
  CFbsBitGc*        iBitTextGc;
};
#endif
