// GDIView.h
//
// Copyright 1999-2000 Sander van der Wal. All rights reserved
//
// $Id: gdiview.h 1.2 2000-09-27 09:37:37+02 svdwal Exp svdwal $

#ifndef GDIVIEW_H
#define GDIVIEW_H

#ifdef __GNUC__
#pragma interface
#endif


// GDI
#ifndef __GDI_H__
#include <gdi.h>
#endif

class TGfxFontSpec;
class TTypefaceSupport;


// GooLib
#ifndef   GTYPES_H
#include "GTypes.h"
#endif
class GString;

// PDF
class FontEncoding;
class GfxFont;


///////////////////////////////////
//
// CGDIFontCache
//
// match GfxFonts with EPOC fonts
//
///////////////////////////////////
struct TGDIFontNameSubst {
  TBuf8<KMaxTypefaceNameLength> iName;
};


class CGDIFontCache: public CBase
{
public:
  CGDIFontCache(CGraphicsDevice* aGd, TGDIFontNameSubst* aFontNameSubst);
  void ConstructL();
  ~CGDIFontCache();

// --o CGDIFontCache
public:
  void Search(GfxFont* aGfxFont, TInt aFontSize, TReal aM[4],
              CFont*& aFont, /*Gushort*/ TUint aMap[], Guchar aRevMap[]);

private:
  void GetGDIFont(GfxFont* aGfxFont, TReal aM[4], 
                  TInt aFontSize, TGfxFontSpec& aEpocFont);
  void FillMaps(GfxFont* aGfxFont, const TGfxFontSpec& aEpocFont,
                /*Gushort*/ TUint aMap[], Guchar  aRevMap[]);

private:
  FontEncoding& iWindowsLatin1Encoding;
  FontEncoding& iSymbolEncoding;
  FontEncoding& iZapfDingbatsEncoding;
  
  CGraphicsDevice* iGd;
  TGDIFontNameSubst* iFontNameSubst;

  CFontCache*      iCache;
};


// RGB color that can store negative values as well (for errors )
class TGDIRgb
{
public:
  inline TGDIRgb(): iRed(0), iGreen(0), iBlue(0) {};
  inline TGDIRgb(TInt aRed, TInt aGreen, TInt aBlue): iRed(aRed), iGreen(aGreen), iBlue(aBlue) {};

  inline TInt Red() const { return iRed; }
  inline TInt Green() const { return iGreen; }
  inline TInt Blue() const { return iBlue; }

  inline void SetRed  (TInt aRed)   { iRed  = aRed; }
  inline void SetGreen(TInt aGreen) { iGreen= aGreen; }
  inline void SetBlue (TInt aBlue)  { iBlue = aBlue; }

  inline TRgb Rgb() const { return TRgb(iRed  < 0 ? 0 : iRed  > 255 ? 255: iRed,
                                        iGreen< 0 ? 0 : iGreen> 255 ? 255: iGreen,
                                        iBlue < 0 ? 0 : iBlue > 255 ? 255: iBlue); }
  inline void SetRgb(const TRgb& aRgb) { iRed  = aRgb.Red();
                                         iGreen= aRgb.Green();
                                         iBlue = aRgb.Blue(); }

  inline TGDIRgb& operator +(TGDIRgb aColor)
  {
    iRed  += aColor.iRed; iGreen+= aColor.iGreen; iBlue += aColor.iBlue;
    return *this;
  };
  friend TGDIRgb operator + (const TGDIRgb& aColor1, const TGDIRgb& aColor2);

private:
  TInt iRed;
  TInt iGreen;
  TInt iBlue;
};

inline TGDIRgb operator + (const TGDIRgb& aColor1, const TGDIRgb& aColor2)
{
  return TGDIRgb(aColor1.iRed  + aColor2.iRed, 
                 aColor1.iGreen+ aColor2.iGreen, 
                 aColor1.iBlue + aColor2.iBlue);
}


//////////////////////////////////
//
// Polygon with Bezier curves
//
//////////////////////////////////
class CPolygon: public CArrayFixFlat<TPoint>
{
public:
  CPolygon(TInt aGranularity);

  TRect BoundingRect() const;
  TBool IsRect() const;

  void AddLineL(const TPoint& aEnd) { AppendL(aEnd); };
  void AddBezierL(const TPoint& aA, const TPoint& aB, const TPoint& aC, const TPoint& aD, TInt aFlatness);
};


//------------------------------------------------------------------------
// CGDIView
//------------------------------------------------------------------------
class CGDIView: public CBase {
public:
  CGDIView() {}

public:
   static TInt FindPixel(const TRgb& aColor, TDisplayMode aDisplayMode);
};


//------------------------------------------------------------------------
// CGDITextView
//------------------------------------------------------------------------
class GfxState;

class CGDITextView: public CGDIView {
public:
  CGDITextView();
  CGDITextView(const CGDITextView& aOriginal);

// --o CGDITextView
public:
  virtual CGDITextView* CopyL();
  
  virtual TInt Rotate() const;

  void UpdateFillColor(const TRgb& aFillColor, CGraphicsContext* aGc);
  void UpdateStrokeColor(const TRgb& aStrokeColor, CGraphicsContext* aGc);
  
  void UpdateRender(TInt aRender, CGraphicsContext* aGc);
  TInt FindTextPixel(TDisplayMode aDisplayMode);
  
  virtual void UpdateFont(GfxState *aState, CGraphicsContext* aGc, CGDIFontCache* aFontCache);
  
  virtual void BeginString(GfxState* aState);
  virtual void EndString(GfxState* aState);
  virtual void DrawChar(CGraphicsContext* aGc, Guchar aChar, TInt aDx);
  
  virtual void SetPos(GfxState* aState);
  virtual void DrawString(GfxState* aState, CGraphicsContext* aGc, GString* aString);
 
protected:
  virtual TInt DevWidth(const TReal& aX, const TReal& aY);
  virtual void DrawSubString(GfxState* aState, CGraphicsContext* aGc, Guchar* aSubString, TInt aNumChars, double aUserWidth, TInt aNumGaps);
  virtual void DrawSpecialChar(CGraphicsContext* aGc, TInt aMappedChar, TInt aDx);

public:
  // speed up for Text drawing: no initialisation in drawChar necessary
  TBuf8<1> iChar;
  TPoint iPos;

protected:
  // text state
  TRgb   iFillColor;
  TRgb   iStrokeColor;
  TInt   iRender;
  CFont* iFont;                 // text font
  double iFontSizeTimesScaling; // text fontsize * text horizontal scaling  

  // map between encodings for the current font.
  /*Gushort*/ TUint iMap[256];    // forward map (PDF code -> font code)
  Guchar  iRevMap[256]; // reverse map (font code -> PDF code)
};


//-----------------------------------------------
//
// CRotatedTextView
//
// Abstract base class to draw rotated characters
//
//-----------------------------------------------
class CWsBitmap;
class CFbsBitmapDevice;
class CFbsBitGc;
class TBitmapUtil;

class CRotatedTextView: public CGDITextView
{
public:
  CRotatedTextView(const CGDITextView& aOriginal, TInt aRotate, CGraphicsDevice* aRotDev, TGDIFontNameSubst* aFontNameSubst);
  void ConstructL();
  ~CRotatedTextView();

protected:
  CRotatedTextView(const CRotatedTextView& aOriginal);

// --o CGDITextView
public:
  TInt Rotate() const;

  void UpdateFont(GfxState *aState, CGraphicsContext* aGc, CGDIFontCache* aFontCache);
 
  void BeginString(GfxState* aState);
  void EndString(GfxState* aState);
  void DrawChar(CGraphicsContext* aGc, Guchar aChar, TInt aDx);
  
  void DrawString(GfxState* aState, CGraphicsContext* aGc, GString* aString);

private:
  TInt DevWidth(const TReal& aX, const TReal& aY);

// --o CRotatedTextView
protected:
   virtual void DrawRotatedStringL(TBitmapUtil* aUtil0, TPoint aZeroEnd) = 0;

private:
  void SetZeroSizes(TInt aNewHeight);
  TPoint UnrotatePos(const TPoint aPos) const;

protected:
  TInt iRotate;

  CGraphicsDevice* iRotDev;
  TGDIFontNameSubst* iFontNameSubst;
  
  TSize             iZeroSizeInPixels;
  TSize             iZeroSizeInTwips;
  CWsBitmap*        iZeroMap;
  CFbsBitmapDevice* iZeroDev;
  CGDIFontCache*    iZeroCache;
  CFbsBitGc*        iZeroGc;

  TPoint iRotStart; 
  TPoint iZeroStart;
};

#endif
