// CPdfPageControl.h
//
// Copyright (c) 1999-2000 Sander van der Wal. All rights reserved.
//
// $Id: pdfpagecontrol.h 1.1 2000-09-17 13:20:25+02 svdwal Exp svdwal $

#ifndef PDFPAGECONTROL_H_
#define PDFPAGECONTROL_H_

// Parent
#ifndef __COECNTRL_H__
#include <coecntrl.h>
#endif
#include "ActivePage.h"          // MActivePageObserver
#include "FbsBitmapOutputDevObserver.h"


//
// state
//
#include "FbsBitmapMemento.h"

struct TPdfPageControlMemento: public TFbsBitmapMemento
{
  TInt iRotate;
  TBool iDoLinks;
};


class CWsBitmap;
class CFbsBitmapOutputDev;


////////////////////////////////////////////////////////////////////////////
//
// class CPdfPageControl
//
////////////////////////////////////////////////////////////////////////////
class CPdfPageControl: public CCoeControl,
                       public MFbsBitmapOutputDevObserver,
                       public MActivePageObserver
{
public:
  CPdfPageControl(PDFDoc* aDoc);
  void ConstructL();
  ~CPdfPageControl();

// -- CCoeControl
public:
  
  TSize MinimumSize() const;
  TSize MaximumSize() const;

private:
  void Draw(const TRect& aRect) const;

// --o MOutputDevObserver
private:
  void NotifyDump();

// --o MActivePageObserver
private:
  void NotifyPageFinishedL(TInt aError);

// -- CPdfPageControl
public:
  void SetModel(PDFDoc* aDoc);

  void SetVisAreaOffset(TPoint aVisAreaOffset);
  TSize PageSizeInPixels();
  void SetSelection(const TRect& aSelection);

  // multitasking drawing of a page
  void StartRenderingL(TInt aPage, TInt aDpi, const TPdfPageControlMemento& aMemento);
  TBool IsRendering();
  void CancelRendering();
  void DisplayLinks(TInt aPage);

  // translate between user coords and device coords (pixels)
  void CvtUserToDev(TReal aX, TReal aY, TPoint& aDev);
  void CvtDevToUser(TPoint aDev, TReal& aX, TReal& aY);
 
private:
  PDFDoc* iDoc;

  TPoint      iVisAreaOffset;
  TRect       iSelection;

  // Bitmap and output device to draw on
  //CFbsBitmap* iPageBitmap;
  CWsBitmap*           iPageBitmap;
  CFbsBitmapOutputDev* iOut;
  
  // the page that is currently being drawn
  CActivePage* iActivePage;
  TInt         iPage;
  TBool        iDoLinks;
  TDesC*       iDrawingMsg;

  // coord transform matrix: 
  TReal iCtm[6]; // user to dev
  TReal iICtm[6]; // dev to user
};


#endif
